// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: ContourMap.h,v 1.6.206.3 2013/05/10 00:04:05 jcorbett Exp $
//
// Find borders and contents of 4-connected regions on a grid.

#ifndef Iv_ContourMap_h
#define Iv_ContourMap_h

#include <vector>
#include <limits>

#include "GridMap.h"
#include "Edge.h"
#include "main.h"
#include "XOS/Portability/export/Port_Type.h"

// R#draw_fence

template <typename Tile>
class ContourMap : GridMap<Tile> {
public:
   // Inheriting from a template class necessitates some extra declarations.
   using GridMap<Tile>::ySize;
   using GridMap<Tile>::xSize;
   using GridMap<Tile>::grid;
   using GridMap<Tile>::Index;
   using GridMap<Tile>::At;
   using GridMap<Tile>::InGrid;
   using GridMap<Tile>::ColumnCount;
   using GridMap<Tile>::RowCount;
   using GridMap<Tile>::MarkPoint;
   using GridMap<Tile>::SweepAllAdjacent;

   ContourMap(Port::UINT32 x, Port::UINT32 y, std::vector<Tile*>& grid)
      : GridMap<Tile>(x, y, grid)
      {
         boundary.reserve(x * y * 4); // One point per perimeter cell.
         mark.resize(grid.size(), false);
         hiddenRows.resize(RowCount(), false);
         hiddenColumns.resize(ColumnCount(), false);
         rowAdjustment.resize(RowCount(), 0);
         columnAdjustment.resize(ColumnCount(), 0);
      }

   virtual ~ContourMap() {}

   // Find the boundary of the region containing (x, y). Also mark
   // tiles in the region to aid in identifying all boundaries.
   bool FindBoundary(Port::UINT32 x, Port::UINT32 y) {
      if (InGrid(x, y)) {
         ++y;
         long duh(RowCount() * ColumnCount() * 4 + 1);

         // Set the starting direction to Up. Consider the color to the
         // right "inside".
         Edge startEdge(Up, x, y);
         Tile* startColor(RightColor(startEdge));

         // Find a point on the left boundary of the region containing (x, y).
         Edge edge(startEdge);
         while (0 < edge.x &&
                LeftColor(edge)->Equal(*RightColor(edge)))
         {
            --edge.x;
            Point p(edge.RightCell());
            MarkCell(p);
         }
         startEdge = edge;
         AddEdge(startEdge);
         // Trace clockwise.
         do {
            Edge straightEdge(edge.GetNext(GoStraight));
            Tile* left(LeftColor(straightEdge));
            Tile* right(RightColor(straightEdge));
            if ((left && left->Equal(*startColor)) &&
                (right && right->Equal(*startColor)))
            {
               edge = edge.GetNext(TurnLeft);
               AddEdge(edge);
            } else if (!(left && left->Equal(*startColor)) &&
                       (right && right->Equal(*startColor)))
            {
               edge = straightEdge;
               ExtendEdge(edge);
            } else {
               edge = edge.GetNext(TurnRight);
               AddEdge(edge);
            }
            MarkRow(edge, *startColor);
         } while (duh-- && !edge.Equal(startEdge));
         ASSERT(duh);
         return true;
      }
      return false;
   }

   bool HasMark(Port::UINT32 i) {
      ASSERT(i < mark.size());
      return mark.at(i);
   }

   // Return the boundary resulting from the last FindBoundary() invocation.
   std::vector<Port::UINT32>& GetBoundary() {
      return boundary;
   }
   void HideColumn(unsigned int x) {
      hiddenColumns.at(x) = true;
   }
   // See HideColumn.
   void HideRow(unsigned int y) {
      hiddenRows.at(y) = true;
   }
   bool IsHiddenColumn(unsigned int x) {
      return hiddenColumns.at(x);
   }
   bool IsHiddenRow(unsigned int y) {
      return hiddenRows.at(y);
   }
   // The hiddenColumns vector will tell us how many columns are
   // hidden before the given column, including the given column. This
   // is accomplished in two phases. In Phase I, hidden columns are
   // set to 1. In Phase II, values are propagated so that every entry
   // has the correct sum.
   //
   // Propagate hidden column values so that every entry tells the
   // number of hidden columns before and including its position.
   void AdjustGrid() {
      {
         ASSERT(!hiddenColumns.empty());
         unsigned int count(hiddenColumns.at(0));
         for (unsigned int i(1); i < hiddenColumns.size(); ++i) {
            count += hiddenColumns.at(i);
            columnAdjustment[i] = count;
         }
      }
      {
         ASSERT(!hiddenRows.empty());
         unsigned int count(hiddenRows.at(0));
         for (unsigned int i(1); i < hiddenRows.size(); ++i) {
            count += hiddenRows.at(i);
            rowAdjustment[i] = count;
         }
      }
   }
   unsigned int GetAdjustedX(unsigned int index) {
      unsigned int x(index % ColumnCount());
      return x - columnAdjustment.at(x);
   }
   unsigned int GetAdjustedY(unsigned int index) {
      unsigned int y(index / ColumnCount());
      return y - rowAdjustment.at(y);
   }
   unsigned int AdjustX(unsigned int x) {
      return x - columnAdjustment.at(x);
   }
   unsigned int AdjustY(unsigned int y) {
      return y - rowAdjustment.at(y);
   }

   // Scan for the next visible row incrementing by dy (usually +/-1).
   // Note: The return value may not be in the grid.
   unsigned int NextVisibleRow(unsigned int x, unsigned int y, int dy) {
      // No need for dy outside this range.
      ASSERT(-2 < dy && dy < 2);
      // It's only ok for dy == 0 if the while loop is never entered.
      ASSERT(dy != 0 || !InGrid(x, y) || !IsHiddenRow(y));
      y += dy;
      while (InGrid(x, y) && IsHiddenRow(y)) {
         y += dy;
      }
      return y;
   }

   // Scan for the next visible row incrementing by dy (usually +/-1).
   // Note: The return value may not be in the grid.
   unsigned int NextVisibleColumn(unsigned int x, unsigned int y, int dx) {
      // No need for dx outside this range.
      ASSERT(-2 < dx && dx < 2);
      // It's only ok for dx == 0 if the while loop is never entered.
      ASSERT(dx != 0 || !InGrid(x, y) || !IsHiddenColumn(x));
      x += dx;
      while (InGrid(x, y) && IsHiddenColumn(x)) {
         x += dx;
      }
      return x;
   }

   // Replaces GridMap::SweepAllAdjacent(). This function should visit
   // every unordered pairs adjacent grid cells exactly once.
   //
   // The nested for loops over (x, y) obviously cover each grid cell
   // exactly once. Each grid cell i has eight neighbors. Four that
   // precede it in the iteration and four that follow it. Since the
   // requirement is to visit unordered pairs once each, it suffices
   // to visit only the neighbors that follow cell i. Since there are
   // hidden rows and columns, the coordinates of the neighbors must
   // be adjusted to skip hidden rows and columns. To show that each
   // following neighbor is visited exactly once, it suffices to show
   // that the next row and column picked when skipping hidden rows
   // and columns is at the minimum possible increment and that each
   // following neighbor for cell i is unique.
   //
   // Clearly x < x2, since x2 is initially x + 1 and only
   // incremented. Likewise for y2. Therefore x != x2 and y != y2.
   //
   // x2 and y2 are incremented a unit at a time and the first
   // non-hidden value terminates the loop.
   //
   // The following neighbors are:
   //   (x2, y), (x2, y2), (x, y2), (xPrev, y2).
   // All obviously distinct.

   void SweepAllAdjacent() {
      unsigned int undefined(std::numeric_limits<unsigned int>::max());
      unsigned int xPrev(undefined);
      for (unsigned int x(0); x < xSize; ++x) {
         if (hiddenColumns.at(x)) {
            continue;
         }
         // Find the next visible (not hidden) column, if any.
         unsigned int x2(x + 1);
         while (x2 < xSize && hiddenColumns[x2]) {
            ++x2;
         }
         for (unsigned int y(0); y < ySize; ++y) {
            if (hiddenRows.at(y)) {
               continue;
            }
            Tile* here(grid[Index(x, y)]);
            if (here) {
               // Find the next visible (not hidden) row, if any.
               unsigned int y2(y + 1);
               while (y2 < ySize && hiddenRows[y2]) {
                  ++y2;
               }
               if (x2 < xSize) {
                  Tile* right(grid[Index(x2, y)]);
                  if (right) {
                     here->Visit(right);
                  }
                  if (y2 < ySize) {
                     Tile* belowRight(grid[Index(x2, y2)]);
                     if (belowRight) {
                        here->Visit(belowRight);
                     }
                  }
               }
               if (y2 < ySize) {
                  Tile* below(grid[Index(x, y2)]);
                  if (below) {
                     here->Visit(below);
                  }
               }
               if (xPrev != undefined && y2 < ySize) {
                  Tile* belowLeft(grid[Index(xPrev, y2)]);
                  if (belowLeft) {
                     here->Visit(belowLeft);
                  }
               }
            }
         }
         xPrev = x;
      }
   }

private:
   void AddEdge(const Edge& edge) {
      boundary.push_back(edge.x);
      boundary.push_back(edge.y);
   }
   void ExtendEdge(const Edge& edge) {
      if (edge.dir == Left || edge.dir == Right) {
         boundary[boundary.size() - 2] = edge.x;
      } else {
         boundary[boundary.size() - 1] = edge.y;
      }
   }
   void MarkCell(Point p) {
      if (InGrid(p.x, p.y)) {
         mark.at(Index(p.x, p.y)) = true;
      }
   }
   Tile* LeftColor(const Edge& edge) {
      Point p(edge.LeftCell());
      return InGrid(p.x, p.y) ? At(p.x, p.y) : 0;
   }
   Tile* RightColor(const Edge& edge) {
      Point p(edge.RightCell());
      return InGrid(p.x, p.y) ? At(p.x, p.y) : 0;
   }
   void MarkRow(Edge& edge, Tile& startColor) {
      Edge e(edge);
      if (e.dir == Up) {
         do {
            Point p(e.RightCell());
            if (InGrid(p.x, p.y)) {
               mark.at(Index(p.x, p.y)) = true;
               ++e.x;
            } else {
               break;
            }
         } while (e.x < ColumnCount() &&
                  RightColor(e) && RightColor(e)->Equal(startColor));
      } else if (e.dir == Down) {
         Edge e(edge);
         while (0 <= e.x &&
                RightColor(e) && RightColor(e)->Equal(startColor))
         {
            Point p(e.RightCell());
            if (InGrid(p.x, p.y)) {
               mark.at(Index(p.x, p.y)) = true;
               --e.x;
            } else {
               break;
            }
         }
      }
   }

private:
   std::vector<bool> mark;
   std::vector<Port::UINT32> boundary;
   std::vector<bool> hiddenColumns;
   std::vector<bool> hiddenRows;
   std::vector<size_t> columnAdjustment;
   std::vector<size_t> rowAdjustment;
};

#endif
