// Copyright (C) 2009-2010 Xilinx, Inc.  All rights reserved
// $Id: Edge.h,v 1.2 2010/11/08 19:03:09 jcorbett Exp $

#ifndef Iv_Edge_h
#define Iv_Edge_h

class Point {
public:
   long x;
   long y;
   Point(long x, long y) : x(x), y(y) {}
};

typedef enum Direction {
   Left = 0, Up = 1, Right = 2, Down = 3
} Direction;

typedef enum Turn {
   TurnRight  = 1,              // 90 deg cw
   GoStraight = 0,
   TurnLeft   = 3               // 3 * 90 deg cw == 90 deg ccw. Using
                                // 3 rather than -1 spares us from
                                // taking the modulus of a negative number.
} Turn;

// Directed line segment that terminates at (x, y) after traveling one
// unit in direction dir. Edges are related to grid cells as defined
// by functions leftCell() and rightCell() below. The left and right
// cells are defined relative to the direction of the edge. The right
// cell of an edge in the Up direction is to the right of the edge,
// but the right cell of an edge in the Down direction is to the left
// of the edge. This simplifies the logic of tracing the border of a
// region.
//
//                      ^
//                      | y+
//                      +---------+---------+
//                      |         | Cell    |
//                      |         | (1, 1)  |
//                      |         |         |
//                      +---------+---------+
//                      ^ Cell    |         |
//       Edge(0, 1, Up) | (0, 0)  |         |
//                      |         |         |
//                      O-------->+---------+--> x+
//                        Edge(1, 0, Right)
//
// Edge(0, 1, Up) starts at Point(0, 0) and ends at Point(0, 1).
// The cell to the right of Edge(0, 1, Up) is Cell(0, 0).
// The cell to the right of Edge(0, 1, Down) is Cell(-1, 0).
class Edge {
public:
   Edge(Direction dir, long x, long y) : dir(dir), x(x), y(y) {}
   bool Equal(const Edge& e) const {
      return e.dir == dir && e.x == x && e.y == y;
   }

   // Three edges start where this one ends. Get the one indicated by
   // the given turn.
   Edge GetNext(Turn turn) const {
      //                   Left     Up  Right   Down
      //                  =====  =====  =====  =====
      const long dx[] = {    -1,     0,     1,     0 };
      const long dy[] = {     0,     1,     0,    -1 };
      Direction nextDir(static_cast<Direction>((dir + turn) % 4));
      return Edge(nextDir, x + dx[nextDir], y + dy[nextDir]);
   }

   // Return the coordinates of the cell to the right of this Edge.
   Point RightCell() const {
      Point p(x, y);
      switch (dir) {
      case Left:                break;
      case Up:           --p.y; break;
      case Right: --p.x; --p.y; break;
      case Down:  --p.x;        break;
      }
      return p;
   }

   // Return the coordinates of the cell to the left of this Edge.
   Point LeftCell() const {
      Point p(x, y);
      switch (dir) {
      case Left:         --p.y; break;
      case Up:    --p.x; --p.y; break;
      case Right: --p.x;        break;
      case Down:                break;
      }
      return p;
   }

public:
   Direction dir;               // direction to (x, y)
   long x, y;                   // end point
};

#endif
