// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: Tile.h,v 1.19.206.15 2013/07/06 22:34:58 jcorbett Exp $
#ifndef Iv_Tile_h
#define Iv_Tile_h

#include <string>
#include <cstring>
#include <vector>
#include <utility>
#include <set>

#include "Classifier.h"
#include "ContourMap.h"
#include "Data.h"
#include "DeviceModel.h"
#include "GroupSetFamily.h"
#include "Databases/Rf/export/Facade/Rf_Cell.h"
#include "Databases/Rf/export/Facade/Rf_Wire.h"

class Tile;
typedef std::set<Tile*> TileSet;
typedef TileSet::iterator TileSetIt;
typedef std::pair<Tile*, Tile*> TileFault;

typedef std::vector<TileFault> TileFaultList;
typedef TileFaultList::iterator TileFaultIt;

typedef std::set<std::pair<size_t, size_t> > PointSet;
typedef PointSet::iterator PointSetIt;

class WireLess {
public:
   bool operator()(const Rf_Wire& a, const Rf_Wire& b) const {
      return strcmp(a.GetName(), b.GetName()) == -1;
   }
};
typedef std::set<Rf_Wire, WireLess> WireSet;
typedef WireSet::iterator WireSetIt;

class XCad_CadNetwork;
typedef std::map<XCad_CadNetwork*, WireSet> NetWireMap;
typedef NetWireMap::iterator NetWireMapIt;

// A tile represents a site-level cell in the device. It contains the
// corresponding device model cell and a pointer to a set of
// groups, owned by a GroupSetFamily instance.
class Tile {
public:
   class TileFaultLess {
   public:
      bool operator()(const TileFault& a, const TileFault& b) const {
         if (a.first->HasGroups() && b.first->HasGroups()) {
            if (GroupSetLess(a.first->GetGroupSet(),
                             b.first->GetGroupSet()))
            {
               return true;
            }
            if (GroupSetLess(b.first->GetGroupSet(),
                             a.first->GetGroupSet()))
            {
               return false;
            }
         }
         int z(strcmp(a.first->cell.GetName(), b.first->cell.GetName()));
         if (z < 0) return true;
         if (0 < z) return false;
         int w(strcmp(a.second->cell.GetName(), b.second->cell.GetName()));
         if (w < 0) return true;
         if (0 < w) return false;
         return false;
      };
   };
   class TileFaultEqual {
   public:
      bool operator()(const TileFault& a, const TileFault& b) const {
         return a.first->cell == b.first->cell &&
            a.second->cell == b.second->cell;
      };
   };
public:
   Tile(Rf_Cell cell,
        TileFaultList& netAdjacencyFaults,
        int& sweepState, GroupSetFamily& gsf,
        bool isUcfMode)
      : cell(cell),
        groupSet(0),
        taintGroupSet(0),
        hasBusMacro(false),
        hasGlobalNet(false),
        hasIsolatedNet(false),
        hasBadNeighbor(false),
        netAdjacencyFaults(netAdjacencyFaults),
        groupSetFamily(gsf),
        sweepState(sweepState),
        isUcfMode(isUcfMode)
   {};
   ~Tile() {};
   void SetCell(Rf_Cell cell);
   Rf_Cell GetCell();
   void AddLogicSite(Rf_Cell logicSite);
   Rf_Cell GetLogicSite();
   bool HasSiteType(const std::string& siteType);
   void SetGroupSet(Data::GroupSet* groupSet);
   Data::GroupSet* GetGroupSet() const;
   Data::GroupSet* GetTaintGroupSet() const;
   bool Equal(const Tile& tile) const;
   void SetHasBusMacro();
   bool HasBusMacro() const;
   void SetHasGlobalNet();
   bool HasGlobalNet() const;
   void SetHasIsolatedNet();
   bool HasIsolatedNet() const;
   bool HasBadContent() const;
   void SetHasBadNeighbor();
   bool HasBadNeighbor() const;
   bool HasViolation() const;
   bool HasProgramming() const;
   void AddTerminal(XCad_CadNetwork* net, Rf_Wire wire);
   void ClearTerminals();
   bool HasTerminal() const;
   bool IsValidFenceTile() const;
   bool HasGroups() const;
   bool ShouldPropagateTo(Tile* adjoined) const;
   const std::string GetName() const;
   const std::string GetLabel() const;
   void AddCompName(const std::string compName);
   void AddNetDescription(const std::string description);
   void AddNet(XCad_CadNetwork* net);
   void CollectNets(Classifier::NameSet& netSet);
   const std::string Describe() const;
   const std::string GetNetsBrief() const;
   void AddNeighbor(Tile* adjoined);
   TileSet& GetNeighbors();
   size_t CountNeighborGroups();
   bool HasTaints() const;
   bool AddTaint(Tile* tile);
   const TileSet& GetTaints() const;
   size_t CountTaints();
   bool HasMultipleTaints();
   bool IsViolation(Tile* adjoined) const;
   void Visit(Tile* adjoined);
public:
   Classifier::NameSet netDescriptions;
   Classifier::NameSet compSet;
   NetWireMap irsPips;
   DeviceModel::CellSet logicSiteSet;
private:
   Rf_Cell cell;
   Rf_Cell logicSite;
   Data::GroupSet* groupSet;
   TileSet neighborSet;
   TileSet taintSet;
   Data::GroupSet* taintGroupSet;
   bool hasBusMacro;
   bool hasGlobalNet;
   bool hasIsolatedNet;
   bool hasBadNeighbor;
   TileFaultList& netAdjacencyFaults;
   GroupSetFamily& groupSetFamily;
   int& sweepState;
   Classifier::NameSet netSet;
   bool isUcfMode;
};

// Add interconnect cells to the given set. See
// CollectSuperSiteTileCoordinates() below.
class AddInts {
public:
   AddInts(PointSet& superSiteSet) : superSiteSet(superSiteSet) {}
   void operator()(Rf_Cell cell) {
      superSiteSet.insert(std::make_pair(cell.GetTileColumn(),
                                         cell.GetTileRow()));
   }
private:
   PointSet& superSiteSet;
};

// Add interconnect cells to the given set. See
// CollectSuperSiteTileCoordinates() below.
class AddCells {
public:
   AddCells(DeviceModel::CellSet& superIntSet) : superIntSet(superIntSet) {}
   void operator()(Rf_Cell cell) {
      superIntSet.insert(cell);
   }
private:
   DeviceModel::CellSet& superIntSet;
};

class ScanForGlobals {
public:
   ScanForGlobals(const boost::shared_ptr<ContourMap<Tile> >& contourMap)
      : hasGlobal(false), contourMap(contourMap)
   {}
   void operator()(Rf_Cell cell) {
      Tile* intTile(contourMap->At(cell.GetTileColumn(),
                                   cell.GetTileRow()));
      hasGlobal |= intTile->HasGlobalNet();
   }
public:
   bool hasGlobal;
private:
   const boost::shared_ptr<ContourMap<Tile> >& contourMap;
};

class AddLogic {
public:
   AddLogic(boost::shared_ptr<ContourMap<Tile> >& contourMap,
            Rf_Cell logicSite)
      : contourMap(contourMap), logicSite(logicSite) {}
   void operator()(Rf_Cell cell) {
      Tile* intTile(contourMap->At(cell.GetTileColumn(),
                                   cell.GetTileRow()));
      intTile->AddLogicSite(logicSite);
   }
private:
   boost::shared_ptr<ContourMap<Tile> >& contourMap;
   Rf_Cell logicSite;
};

// Get the union of all the groups associated with a collection of cells.
class GatherGroups {
public:
   GatherGroups(boost::shared_ptr<ContourMap<Tile> >& contourMap,
                GroupSetFamily& groupSetFamily)
      : contourMap(contourMap),
        groupSetFamily(groupSetFamily),
        groupSet(0)
      {}
   void operator()(Rf_Cell cell) {
      Tile* intTile(contourMap->At(cell.GetTileColumn(),
                                   cell.GetTileRow()));
      if (intTile) {
         groupSet =
            groupSetFamily.FindOrMakeUnion(groupSet, intTile->GetGroupSet());
      }
   }
   Data::GroupSet* GetGroupSet() {
      return groupSet;
   }
private:
   boost::shared_ptr<ContourMap<Tile> >& contourMap;
   GroupSetFamily& groupSetFamily;
   Data::GroupSet* groupSet;
};

// R#check_irs_separation, R#check_direct_routing, R#check_logic_tiles -
// Propagate group information from the given site to the interconnect
// tiles that it is directly connected to.
//
// The groups occupied by these tiles should be shared across all
// these tiles because the logic site connects them.
class MarkOneComp {
public:
   MarkOneComp(boost::shared_ptr<ContourMap<Tile> >& contourMap,
               Rf_Cell logicSite,
               Data::GroupSet* groupSet,
               GroupSetFamily& groupSetFamily)
      : contourMap(contourMap),
        logicSite(logicSite),
        siteName(logicSite.GetName()),
        groupSet(groupSet),
        groupSetFamily(groupSetFamily)
      {}
   void operator()(Rf_Cell cell) {
      Tile* intTile(contourMap->At(cell.GetTileColumn(),
                                   cell.GetTileRow()));
      if (intTile) {
         intTile->AddCompName(siteName);
         intTile->AddLogicSite(logicSite);
         if (isUcf) {
            intTile->SetHasIsolatedNet();
         }
         groupSet =
            groupSetFamily.FindOrMakeUnion(groupSet, intTile->GetGroupSet());
         intTile->SetGroupSet(groupSet);
      } else {
         Warn(121, "No tile for site " + siteName + ".");
      }
   }
   void SetUcfMode() { isUcf = true; }
private:
   boost::shared_ptr<ContourMap<Tile> >& contourMap;
   Rf_Cell logicSite;
   const std::string siteName;
   Data::GroupSet* groupSet;
   GroupSetFamily& groupSetFamily;
   bool isUcf;
};

const std::string SetToString(const Data::GroupSet* gs,
                              const std::string separator = ", ");

#endif
