// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: Classifier.c,v 1.8.206.2 2013/04/18 17:48:35 jcorbett Exp $

#include "JDBHelpers/Pds/export/Pdl/Pds_PdlBlock.h"

#include "main.h"

#include "Classifier.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.8.206.2 $";

// R#report_interregion_signals - Is the given block one that was
// generated for automatically inserted DCM autocalibration logic?
//
// This block is associated with clock management.
//
// This logic is used to prevent Negative-Bias Temperature Instability
// (NBTI) effects inserted by default in Virtex-4 designs for unused
// Digital Clock Manager resources.
bool Classifier::IsDcmBlock(XCad_CadBlock* block) {
   if (block) {
      Xdm_Property* p(block->GetPropertyByName("LIT_NON_USER_LOGIC"));
      if (p && p->GetPropertyType() == Xdm_Property::XDM_PROPERTY_STRING) {
         string pv(p->GetPropertyString());
         return pv == "DCM_PMV_FEATURE" || pv == "DCM_STANDBY";
      } else {
         return string(block->GetName()) == "DCM_AUTOCALIBRATION_TOP_DCM_ADV";
      }
   }
   return false;
}

// R#report_interregion_signals
// Non-specific non-user logic.
bool Classifier::IsNonUserBlock(XCad_CadBlock* block) {
   return block && !!block->GetPropertyByName("LIT_NON_USER_LOGIC");
}

// R#report_interregion_signals
bool Classifier::IsInsideBusMacro(XCad_CadNetwork* net) {
   ASSERT(net);
   string netName(net->GetReportableName());
   return busMacroInternalNets.find(netName) != busMacroInternalNets.end();
}

// R#report_interregion_signals
bool Classifier::IsAttachedToBusMacro(XCad_CadNetwork* net) {
   ASSERT(net);
   string netName(net->GetReportableName());
   return busMacroAttachedNets.find(netName) != busMacroAttachedNets.end();
}

// R#report_interregion_signals
bool Classifier::IsDrivenByRegionalClockBuffer(XCad_CadNetwork* net) {
   bool result(false);
   const string BUFR("BUFR");
   const string BUFMR("BUFMR");
   const string BUFIO("BUFIO");
   XCad_CadNetwork::Signals sigs(net->GetSignals());
   for (XCad_CadNetwork::Signals::iterator sigIt(sigs.begin());
        sigIt != sigs.end();
        ++sigIt)
   {
      XCad_CadSignal::Pins pins((*sigIt)->GetPins());
      for (XCad_CadSignal::Pins::iterator pinIt(pins.begin());
           pinIt != pins.end();
           ++pinIt)
      {
         XCad_CadPin* pin(*pinIt);
         if (pin->IsSignalDriver()) {
            XCad_CadBlock* block(pin->GetParent());
            const Pds_PdlBlock* typeBlock(block->GetTypeBlock());
            if (typeBlock) {
               const string typeBlockName(typeBlock->GetName());
               if ((BUFR == typeBlockName) ||
                   (BUFMR == typeBlockName) ||
                   (BUFIO == typeBlockName))
               {
                  result = true;
                  break;
               }
            }
         }
      }
   }
   return result;
}

// The global buffers for Virtex-4 and Virtex-5 can be found in the Virtex-4
// Libraries Guide for HDL Designers and the Virtex-5 Libraries Guide
// for HDL Designers, respectively, on the Xilinx website. They are
// substantially similar and consistent.
//
// The primitive names for all global clock buffers have the prefix
// "BUFG". All primitive names with this prefix are global clock
// buffers.
//
// The primitive names for all digital clock managers and phase locked
// loop clock circuits have either the prefix "DCM" or "PLL". All
// primitive names with these prefix are global clock sources, however
// not all outputs of these primitives are clock signals. All and only
// the output pins names with the prefix "CLK" are clock drivers.
bool Classifier::IsDrivenByGlobalClock(XCad_CadNetwork* net) {
   bool result(false);
   XCad_CadNetwork::Signals sigs(net->GetSignals());
   for (XCad_CadNetwork::Signals::iterator sigIt(sigs.begin());
        sigIt != sigs.end();
        ++sigIt)
   {
      XCad_CadSignal::Pins pins((*sigIt)->GetPins());
      for (XCad_CadSignal::Pins::iterator pinIt(pins.begin());
           pinIt != pins.end();
           ++pinIt)
      {
         XCad_CadPin* pin(*pinIt);
         if (pin->IsSignalDriver()) {
            XCad_CadBlock* block(pin->GetParent());
            const Pds_PdlBlock* typeBlock(block->GetTypeBlock());
            if (typeBlock) {
               const char* typeName(typeBlock->GetName());
               ASSERT(typeName);
               // if BUFG is a prefix of typeName...
               if (IsPrefix("BUFG", typeName)) {
                  // The only output of any of the BUFGs is a clock.
                  result = true;
                  break;
               }
               if (IsPrefix("IBUFDS", typeName)) {
                  result = true;
                  break;
               }
               if (IsPrefix("BUFH", typeName)) {
                  result = true;
                  break;
               }
               // if PMCD matches typeName exactly...
               if (strcmp("PMCD", typeName) == 0) {
                  // All outputs of the Phased-Matched Clock Divider
                  // are clocks.
                  result = true;
                  break;
               }
               const char* pinName(pin->GetName());
               // if DCM is a prefix of typeName and CLK is a prefix
               // of the driver pin...
               if (IsPrefix("DCM", typeName) && IsPrefix("CLK", pinName)) {
                  result = true;
                  break;
               }
               // if PLL is a prefix of typeName and CLK is a prefix
               // of the driver pin...
               if (IsPrefix("PLL", typeName) && IsPrefix("CLK", pinName)) {
                  result = true;
                  break;
               }
               if (IsPrefix("MMCM", typeName) && IsPrefix("CLK", pinName)) {
                  result = true;
                  break;
               }
            }
         }
      }
   }
   return result;
}

// R#report_interregion_signals - Get the blocks that directly contain
// nets marked with the IS_BUS_MACRO property. Collect all the
// nets inside bus macros and all the nets attached to bus
// macros.
void Classifier::CollectBusMacros(CadFile::NetVec& nets) {
   // Collect all the blocks containing bus macro nets.
   for (CadFile::NetVecIt netIt(nets.begin()); netIt != nets.end(); ++netIt) {
      XCad_CadNetwork* net(*netIt);
      if (net->GetPropertyByName("IS_BUS_MACRO")) {
         // Record which groups share this bus macro net.
         XCad_CadSignal* topSignal(net->GetTopSignal());
         ASSERT(topSignal);
         XCad_CadBlock* busMacroBlock(topSignal->GetParent());
         ASSERT(busMacroBlock);
         busSet.insert(busMacroBlock);
      }
   }
   // Collect all the nets attached to blocks containing bus macro nets.
   for (BlockSetIt bsIt(busSet.begin()); bsIt != busSet.end(); ++bsIt) {
      XCad_CadBlock* macroBlock(*bsIt);
      typedef XCad_CadBlock::Pins PinVec;
      PinVec pins(macroBlock->GetPins());
      for (PinVec::iterator pinIt(pins.begin()); pinIt != pins.end(); ++pinIt)
      {
         XCad_CadPin* pin(*pinIt);
         XCad_CadSignal* signal(pin->GetSignal());
         if (signal) {
            XCad_CadNetwork* net(signal->GetNetwork());
            ASSERT(net);
            NetName netName(net->GetReportableName());
            busMacroAttachedNets.insert(netName);
         }
      }
   }
   // Collect all the nets internal to blocks containing bus macro nets.
   for (BlockSetIt bsIt(busSet.begin()); bsIt != busSet.end(); ++bsIt) {
      XCad_CadBlock* block(*bsIt);
      XCad_CadNetwork::Signals sigs(block->GetSignals());
      for (XCad_CadNetwork::Signals::iterator sigIt(sigs.begin());
           sigIt != sigs.end();
           ++sigIt)
      {
         XCad_CadNetwork* net((*sigIt)->GetNetwork());
         NetName netName(net->GetReportableName());
         if (busMacroAttachedNets.find(netName) ==
             busMacroAttachedNets.end())
         {
            busMacroInternalNets.insert(netName);
         }
      }
   }
}

Classifier::BlockSetIt Classifier::BusSetBegin() {
   return busSet.begin();
}

bool Classifier::BusSetItValid(const Classifier::BlockSetIt& bsIt) {
   return bsIt != busSet.end();
}

// Headings used reporting nets in categories. Note that a net can
// logically belong to multiple categories; but a single category is
// sufficient to understand the role of the net in isolation analysis.
const Classifier::CodeHeading netCategories[] = {
   {Classifier::NoClassification,    ""},
   {Classifier::InterRegionSignal,   "Uncategorized User Global Nets"},
   {Classifier::Ground,              "Nets Driven By Ground"},
   {Classifier::Vcc,                 "Nets Driven By Vcc"},
   {Classifier::DriverConstant,
    "Nets Driven By Constants or Unused Blocks"},
   {Classifier::DedicatedRoute,
    "Nets Without General Routing (Including Clock Nets)\n"
    "\n"
    "  Each of the following nets is either internal\n"
    "  to a logic block (such as a CLB or an IOB) or is\n"
    "  part of a global clock net, therefore they do\n"
    "  not need to be considered for isolation analysis."},
   {Classifier::SourceGlobalClock,
    "Nets Driven by Global Clock Sources"
    " (BUFG, DCM, PLL, and PMCD)"},
   {Classifier::InsideBusMacro,      "Nets Inside Trusted Bus Macros"},
   {Classifier::AttachedToBusMacro,  "Nets Attached to Trusted Bus Macros"},
   {Classifier::SourceRegionalClock,
    "Nets Driven by Regional Clock Buffers"},
   {Classifier::NonUserDcm,
    "Non-user Digital Clock Manager Nets\n"
    "\n"
    "  The following nets are associated with the Digital\n"
    "  Clock Manager (DCM) negative bias temperature instability\n"
    "  (NBTI) mitigation. These non-user nets are "
    "clock-related\n"
    "  and therefore do not require isolation."},
   {Classifier::NonUser1,            "Non-user Nets"},
   {Classifier::NonUser2,            "Non-user Nets"}
};

// R#report_interregion_signals - Some nets do not need to be used
// as starting or ending points in the search for isolation
// violations. This function gives section headings for verbose
// analysis of nets that appear in more than one isolation group.
Classifier::NetCode Classifier::Classify(XCad_CadNetwork* net, XCad_RMed* rmed)
{
   XCad_RMedRug* rug(rmed->GetRug(net));
   NetCode reason(NoClassification);
   NetName netName(net->GetReportableName());
   if (net->IsGround()) {
      reason = Ground;
   } else if (net->IsVcc()) {
      reason = Vcc;
   } else if (net->IsConstant()) {
      reason = DriverConstant;
   } else if (!rug) {
      reason = DedicatedRoute;
   } else if (IsDrivenByGlobalClock(net)) {
      reason = SourceGlobalClock;
   } else if (IsInsideBusMacro(net)) {
      reason = InsideBusMacro;
   } else if (IsAttachedToBusMacro(net)) {
      reason = AttachedToBusMacro;
   } else if (IsDrivenByRegionalClockBuffer(net)) {
      reason = SourceRegionalClock;
   } else {
      // Classify nets based on block properties.
      XCad_CadSignal* topSignal(net->GetTopSignal());
      XCad_CadBlock* sigBlock(topSignal->GetParent());
      if (IsDcmBlock(sigBlock)) {
         reason = NonUserDcm;
      } else if (IsNonUserBlock(sigBlock)) {
         reason = NonUser1;
      } else {
         bool classified(false);
         typedef XCad_CadBlock::Pins PinVec;
         PinVec pins(topSignal->GetPins());
         for (PinVec::iterator pinIt(pins.begin());
              pinIt != pins.end() && !classified;
              ++pinIt)
         {
            XCad_CadPin* topPin(*pinIt);
            XCad_CadBlock* cadBlock(topPin->GetParent());
            typedef XCad_CadBlock::Blocks CadBlocks;
            CadBlocks kids(cadBlock->GetBlocks());
            for (CadBlocks::iterator blockIt(kids.begin());
                 !classified && blockIt != kids.end();
                 ++blockIt)
            {
               XCad_CadBlock* kid(*blockIt);
               if (IsDcmBlock(kid)) {
                  reason = NonUserDcm;
                  classified = true;
               } else if (IsNonUserBlock(kid)) {
                  reason = NonUser2;
                  classified = true;
               }
            }
         }
      }
   }
   ASSERT(netCategories[reason].code == reason);
   return reason;
}

// Fetch the heading for a given net.
const string Classifier::GetHeading(NetCode reason) {
   ASSERT(netCategories[reason].code == reason);
   return netCategories[reason].heading;
}

// Fetch the heading for a given net.
const string Classifier::GetHeading(XCad_CadNetwork* net, XCad_RMed* rmed) {
   NetCode reason(Classify(net, rmed));
   ASSERT(netCategories[reason].code == reason);
   return netCategories[reason].heading;
}

// Return the internal name of the given bus macro block. This
// information may be useful for spotting an accidental use of a bus
// macro that isn't approved for SCC designs.
const string Classifier::GetBusMacroSource(const XCad_CadBlock* macroBlock) {
   string source("<unknown>");
   const Xdm_Property*
      xdmProp(macroBlock->GetPropertyByName("Xda_HierMgr_SrcRef"));
   if (xdmProp && xdmProp->GetPropertyObject()) {
      source = xdmProp->GetPropertyObject()->GetName();
   }
   return source;
}

