// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: DeviceModel.c,v 1.24.206.17 2013/07/11 01:02:35 jcorbett Exp $

#include "TUtilities/export/regex.hpp"

#include "Databases/PersonalityModule/export/baspmpic.h"
#include "Databases/Rf/export/Facade/Rf_DeviceMgr.h"
#include "Databases/Rf/export/Facade/Rf_Device.h"
#include "Databases/Rf/export/Facade/Rf_Package.h"
#include "Databases/Rf/export/Facade/Rf_CellContainer.h"
#include "Databases/Rf/export/Facade/Rf_Grid.h"
#include "Utilities/export/UtilSharedObject.h"

#include "DeviceModel.h"
#include "device.inc"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.24.206.17 $";

// To map pin names to IOB coordinates, we'll need an Rf_Package.
// To get an Rf_Package, we'll need an Rf_Device. To get an
// Rf_Device, we'll need a PM_PIC.
//
// N.B. This helper function is needed because there are two
// very different constructors for this class. One is given
// a const Rf_Device pointer. The other must call several functions
// before an Rf_Device pointer is available. Since the Rf_Device
// pointer is const, we use this function to return a fully
// initialized PM_PIC instance from which we can obtain a const
// Rf_Device pointer.
PM_PIC* NewPic(const string deviceName, const string packageName) {
   SuppressLibraryMessages suppressor;

   PM_PIC* pmpic(new PM_PIC());
   // Before we can validate the device name, we need to know
   // the architecture to check the device name against, but there are
   // so many architecture names that it's easier to generate the
   // entire list off-line and just check them all. At compile time,
   // the script gendevlist.bash is used to generate the file
   // device.inc.

   string arch;
   size_t deviceCount(sizeof archDeviceList / sizeof archDeviceList[0]);
   for (size_t i(0); i < deviceCount; i++) {
      if (deviceName == archDeviceList[i].device) {
         arch = archList[archDeviceList[i].archIndex];
      }
   }
   if (!arch.empty()) {
      pmpic->set_arch(arch.c_str());
   } else {
      FatalUsage(3, "Device " + deviceName + " was not found.\n"
                 "Check the spelling of the -device parameter.");
   }
   // R#report_errors
   // Now that a device list is generated at compile time, this
   // is a sanity check only.
   if(!pmpic->isvaliddevicename(deviceName.c_str(), TRUE)) {
      FatalUsage(3, "Device " + deviceName + " was not found.\n"
                 "Check the spelling of the -device parameter.");
   }
   pmpic->set_device(deviceName.c_str());
   // R#report_errors
   if (!pmpic->isvalidpackagename(packageName.c_str(), TRUE)) {
      FatalUsage(14, "Package " + packageName + " was not found.\n"
                 "Check the spelling of the -package parameter.");
   }
   pmpic->set_package(packageName.c_str());
   return pmpic;
}

// Get the Rf_TilePinHelpers instance, building it if necessary.  This
// object is used to find the interconnect tile or tiles associated
// with a site. It does not appear to work in all cases, but it works
// for tiles that the user can use as fence tiles.
Rf_TilePinHelpers* GetHelpers(const Rf_Device* device) {
   Rf_TilePinHelpers*
      helpers(Util::SharedObject<Rf_TilePinHelpers>::Key(device).Get());
   if (!helpers) {
      helpers =
         Util::SharedObject<Rf_TilePinHelpers>::Key(device).Create(device);
   }
   helpers->Build();
   return helpers;
}

const string SetToString(const DeviceModel::CellSet cs,
                         const string separator)
{
   string result;
   string sep;
   for (DeviceModel::CellSet::const_iterator csIt(cs.begin());
        csIt != cs.end();
        ++csIt)
   {
      const Rf_Cell cell(*csIt);
      result += sep + string(cell.GetName());
      sep = separator;
   }
   return result;
}

// Load a device model given the device name and package name.
DeviceModel::DeviceModel(const string deviceName, const string packageName)
   : pmpic(NewPic(deviceName, packageName)), isPicOwner(true),
     device(Rf_DeviceMgr::AcquireDevice(pmpic)),
     package(device->GetPackageData()),
     helpers(GetHelpers(device))
{}

// Hold a device model that was already loaded as a side-effect
// of loading a placed and routed design.
DeviceModel::DeviceModel(const Rf_Device* device)
   : pmpic(device->GetPmPic()), isPicOwner(false),
     device(device),
     package(device->GetPackageData()),
     helpers(GetHelpers(device))
{}

DeviceModel::~DeviceModel() {
   // Only release resources we own.
   if (helpers) {
      Util::SharedObject<Rf_TilePinHelpers>::Key(device).Release(helpers);
      helpers = 0;
   }
   if (isPicOwner && device) {
      Rf_DeviceMgr::ReleaseDevice(device);
      device = 0;
      // The Rf module deletes the pmpic. We gave it ownership in
      // AcquireDevice.
   }
   pmpic = 0;
   package = 0;
}

const Rf_Device* DeviceModel::GetRfDevice() const {
   return device;
}

const Rf_Package* DeviceModel::GetRfPackage() const {
   return package;
}

// R#report_provenance -
// Return the part name. For example, "xc5vfx30t".
const string DeviceModel::GetPartName() const {
   return string(pmpic->getDatabookName());
}

// R#accept_isolation_regions -
// Return the architecture. For example, "virtex4".
const string DeviceModel::GetArchitectureName() const {
   return string(pmpic->get_arch());
}

// Return the number of columns on the package.
Port::UINT32 DeviceModel::GetPackageColumnCount() const {
   Port::INT32 columns(package->GetColumnCount());
   ASSERT(0 < columns);
   return static_cast<Port::UINT32>(columns);
}

// Return the number of rows on the package.
Port::UINT32 DeviceModel::GetPackageRowCount() const {
   Port::INT32 rows(package->GetRowCount());
   ASSERT(0 < rows);
   return static_cast<Port::UINT32>(rows);
}

// Return the number of columns of tiles on the device.
Port::UINT32 DeviceModel::GetTileColumnCount() const {
   Port::UINT32 columns(device->GetBaseGrid()->GetColumnCount());
   return columns;
}

// Return the number of rows of tiles on the device.
Port::UINT32 DeviceModel::GetTileRowCount() const {
   Port::UINT32 rows(device->GetBaseGrid()->GetRowCount());
   return rows;
}

// Return all the tile cells in the device (whether configured or
// not).
Rf_CellContainer DeviceModel::GetCells() const {
   return device->GetTopCellDefTile().GetCells();
}

// Return all the site cells in the device (whether configured or
// not).
Rf_CellContainer DeviceModel::GetSites() const {
   return device->GetTopCellDef().GetCells();
}

// Get the package pin instance for a given cell.
Rf_PackagePin DeviceModel::GetPackagePin(Rf_Cell cell) const {
   return package->GetPackagePin(cell);
}

// Return the tile that contains the given site.
Rf_Cell DeviceModel::GetSiteParentTile(Rf_Cell site) const {
   ASSERT(site);
   Rf_CellDef tileDef(device->GetTopCellDefTile());
   return tileDef.GetCell(site.GetTileIndex());
}

// R#define_fence - Is the site name part of a tile that needs to be
// analyzed?  This is the list of site types from the isolation
// methodology white paper.
const boost::regex userTilePat("DCM.*|DSP48.*|EMAC|GT11|GTP.*|GTX.*|IOB.*|MCB|"
                               "PCIE|PCIE_A1|PCIE_2_1|PLL.*|PPC.*|PS7.*|"
                               "RAMB.*|SLICE.*|MMCM.*|PHASER.*|PHY_CONTROL|"
                               "IN_.*|OUT_.*|PMV.*|IDELAY|ILOGIC|OLOGIC");

// Does this site type need to be analyzed with respect to isolation?
bool DeviceModel::IsUserTileType(const string& siteType) const {
   return boost::regex_match(siteType, userTilePat);
}

// Is this site configured in such a way that we cannot rule out the
// possibility that it causes some of its neighbors to be shorted
// together?
//
// The PCIE_2_1 is called out specifically because it the only 7
// series supersite that's unconditionally treated as providing no
// fence between its neighbors.
//
// Ditto for the Zynq PSS!
bool DeviceModel::IsShortableSuperSite(const Rf_Cell superSite) const {
   const string siteType(superSite.GetCellDef().GetName());
   return siteType == "PCIE_2_1" || siteType == "PS7";
}


// All the sites below connect without branching to interconnects from
// a single column, and no other sites connect directly to
// interconnects in this column. The pattern below us used for several
// purposes: 
//    * Identify the config center column interconnects
//    * Verify that the config center sites use only one column of
//      interconnects, and
//    * Verify that no other sites use these interconnects.
const boost::regex DeviceModel::configCenterPat("BSCAN|CAPTURE|DCIRESET|"
                                                "DNA_PORT|EFUSE_USR|"
                                                "FRAME_ECC|ICAP|STARTUP|"
                                                "USR_ACCESS|XADC");

// Several sites are grouped in an area called the Config Center
// area. This area contains special-purpose sites related to
// configuration, SEU checking, device settings, and analog-to-digital
// conversion. The config center area is treated similarly to a
// supersite, however there is no distinguished site associated with
// the area so special case code is required.
bool DeviceModel::IsConfigCenterType(const std::string& siteType) const {
   return boost::regex_match(siteType, configCenterPat);
}

// For 7 series, the clock buffer column cannot be used as a fence.
// This pattern is used to identify sites of the clock buffer column.
const boost::regex DeviceModel::clockBufferPat("BUFG|BUFHCE|MTBF2|BUFGCTRL|"
                                               "GCLK_TEST_BUF");

// The 7 series contains columns of clock management tiles (CMTs). The
// pattern below identifies these.
const boost::regex clockColumnPat("PHASER.*|PHY_CONTROL|MMCM.*|IN_.*|OUT_.*|"
                                  "PLL.*|PMV.*");

// Is the given site in a 7 series CMT column?
bool DeviceModel::Is7SeriesClockSite(Rf_Cell site) const {
   const std::string siteType(site.GetCellDef().GetName());
   return boost::regex_match(siteType, clockColumnPat);
}

// Some connections to the 7 series phaser sites have been shown by
// analysis to not need further automated isolation analysis. These
// are typically clock and status signals that do not contain critical
// program information.
const boost::regex phaserSafePinPat("CTSBUS[01]|DQSBUS[01]|DTSBUS[01]|"
                                    "[IO]CLKDIV|[IO]CLK|OCLKDELAYED|"
                                    "PHASEREFCLK|PHYCTLEMPTY|"
                                    "PHYCTLMSTREMPTY|RDENABLE|WRENABLE");

// Can this pin be considered safe according to device schematic
// analysis and thus excluded from further isolation analysis?
bool DeviceModel::IsPhaserSafePin(Rf_Pin pin) const {
   const string pinName(pin.GetName());
   return boost::regex_match(pinName, phaserSafePinPat);
}

// Pattern matching Spartan-6 interconnect tile types
const boost::regex spartan6IntTilePat(
   "INT|INT_TERM|INT_TERM_BRK|LIOI_INT|LIOI_INT_BRK|"
   "INT_GCLK|IOI_INT|INT_BRAM|INT_BRK|INT_BRAM_BRK"
   );

// R#define_fence - Is cell an interconnect tile?
bool DeviceModel::IsInterconnectCell(const Rf_Cell cell) const {
   if (!cell) {
      FatalInternal(97, "Invalid cell passed to IsInterconnectCell.");
   }
   if (!cell.IsTileLevel()) {
      FatalInternal(98, "Non-tile cell passed to IsInterconnectCell.");
   }
   bool result(false);
   const string cellTypeName(cell.GetCellDef().GetName());
   if (IsSpartan6()) {
      // Spartan-6
      result = boost::regex_match(cellTypeName, spartan6IntTilePat);
   } else if (IsVirtex4()) {
      result = (cellTypeName == "INT") ||
         (cellTypeName == "INT_SO") ||
         (cellTypeName == "INT_SO_DCM0") ||
         (cellTypeName == "L_TERM_INT");
   } else if (Is7Series()) {
      result = (cellTypeName == "INT_L") || (cellTypeName == "INT_R");
   } else if (IsVirtex5()) {
      result = (cellTypeName == "INT");
   } else {
      // To provide preliminary support for future devices, assume
      // interconnect tiles at least have type "INT". Supported 
      // architectures are checked elsewhere.
      result = (cellTypeName == "INT");
   }
   return result;
}

// Given the site name, find the corresponding site.
const Rf_Cell DeviceModel::GetSite(const Rf_CellDef& topCellDef,
                                   const string siteName) const
{
   // Find the tile for the named slice. To do this, iterate over the
   // slices, looking for slices that match the UCF slice coordinates,
   // then find the tile that contains each slice.
   Rf_CellContainer sites(topCellDef.GetCells());
   Rf_Cell result;
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      Rf_Cell site(*sitesIt);
      string candidateName(site.GetName());
      if (siteName == candidateName) {
         result = site;
      }
   }
   if (!result) {
      // In UCF mode it is possibly a user input error. In NCD mode
      // this is a fatal internal error.
      FatalInput(101, "Unable to find user tile " + siteName + ".");
   }
   return result;
}

// FPGA components spanning multiple grid cells are called
// supersites. We identify them here as sites that use multiple
// interconnect tiles.
bool DeviceModel::SpansMultipleTiles(const Rf_Cell site) const {
   return helpers->SpansMultipleInterconnectTiles(site);
}

// Get a top or bottom interconnect tile for a site.
// Note: not all sites have associated interconnects. If there are
// none, this function returns the invalid cell.
Rf_Cell DeviceModel::GetIntTile(const Rf_Cell site,
                                const Rf_CellDef& tileDef,
                                int directionSign) const
{
   Rf_Cell result;              // Initially the invalid cell.
   ASSERT(directionSign == 1 || directionSign == -1);
   bool isGreater(directionSign == 1);

   BoundInts bb;
   MapIntTiles(bb, site, tileDef);
   if (bb.bottom && bb.top) {
      result = isGreater ? bb.bottom : bb.top;
   }
   return result;
}

// Given a single rectangle from an area range constraint, find the
// cells that correspond to the top left- and bottom right
// corners. The cells are used to obtain tile coordinates in a single
// universal coordinate system. Each type of site range (SLICE, IOB,
// etc.) has its own coordinate system.
void DeviceModel::GetAreaRangeCorners(SiteRect& siteRect) {
   Rf_CellDef tileDef(device->GetTopCellDefTile());
   if (!tileDef) {
      FatalInternal(106, "Tile cell definition is not valid.");
   }
   boost::smatch result;
   // Match two strings separated by a colon. A very cute regex.
   const boost::regex rangePat("(.*):(.*)");
   bool gotMatch(boost::regex_match(siteRect.rangeString, result, rangePat));
   ASSERT(gotMatch);

   Rf_CellDef topCellDef(device->GetTopCellDef());
   ASSERT(!!topCellDef);

   Rf_Cell site1(GetSite(topCellDef, result.str(1)));
   Rf_Cell site2(GetSite(topCellDef, result.str(2)));

   // If the corners are swapped in the UCF, swap them back here.  In
   // canonical form site x1 <= x2 and site y1 <= y2. We may get the
   // pairs in either order. We may get the 'backslash diagonal' (\)
   // rather than the canonical slash diagonal (/). Note: tile row
   // coordinates increase downwards, but site coordinates in the UCF
   // range text increase upward.
   Rf_Cell p1(GetSiteParentTile(site1));
   Rf_Cell p2(GetSiteParentTile(site2));
   Port::UINT32 r1(p1.GetTileRow());
   Port::UINT32 r2(p2.GetTileRow());
   if (r2 <= r1) {
      siteRect.lowerCell = p1;
      siteRect.upperCell = p2;
   } else {
      siteRect.lowerCell = p2;
      siteRect.upperCell = p1;

      Rf_Cell temp(site1);
      site1 = site2;
      site2 = temp;
   }
   siteRect.canonical =
      siteRect.lowerCell.GetTileColumn() <= siteRect.upperCell.GetTileColumn();
   ASSERT(siteRect.upperCell.GetTileRow() <= siteRect.lowerCell.GetTileRow());
   ASSERT(siteRect.lowerCell.GetTileColumn()
          <= siteRect.upperCell.GetTileColumn()
          == siteRect.canonical);

   CornerInts ci1;
   MapIntTiles(ci1, site1, tileDef);
   siteRect.lowerInt = ci1.lowerLeft;
   ASSERT(siteRect.lowerInt);

   if (site1 != site2) {
      CornerInts ci2;
      MapIntTiles(ci2, site2, tileDef);
      siteRect.upperInt = ci2.upperRight;
   } else {
      siteRect.upperInt = ci1.upperRight;
   }
   ASSERT(siteRect.upperInt);
}

// There are lots of site types. An informal survey of Virtex-4,
// Virtex-5, and Spartan-6 yielded the following:
//    BSCAN, BUFDS, BUFGCTRL, BUFGMUX, BUFH, BUFIO, BUFIO2, BUFIO2FB,
//    BUFR, CRC32, CRC64, DCI, DCM, DCM_ADV, DSP48, FIFO16, GT11,
//    GTPA1_DUAL, ICAP, IDELAYCTRL, ILOGIC, IOB, IODELAY, IPAD, MCB,
//    MULT18X18, OLOGIC, OPAD, PCIE, PLL_ADV, PLL_ADV, PMVBRAM, PPC405,
//    PPC405_ADV, PPC440, RAMB16, RAMB36, RAMB36, RAMB8, SLICE, SYSMON,
//    TBUF, TEMAC.
//
// Rather than try to track the list of site types as new device
// models come along, IVT will just accept anything that looks like
// a range.
//
// Typical value: SLICE_X0Y110:SLICE_X1Y127
// Match results:
//   result.str(0) = "SLICE_X0Y110:SLICE_X1Y127"
//   result.str(1) = "SLICE"
//   result.str(2) = "0"
//   result.str(3) = "110"
//   result.str(4) = "SLICE"
//   result.str(5) = "1"
//   result.str(6) = "127"

const string sp("\\s*");           // match optional whitespace
const string site("([A-Z0-9_]+)"); // match a site name
const string num("(\\d+)");        // a non-negative integer
const boost::regex
areaRangePat(sp
             + site + "_X" + num + "Y" + num + sp
             + ":" + sp
             + "\\1_X" + num + "Y" + num + sp,
             boost::regex::normal | boost::regbase::icase);

const boost::regex
sitePat(sp + site + "_X" + num + "Y" + num + sp);

// Parse a range string matching areaRangePat.
// Later refactor this to handle single sites.
bool DeviceModel::ParseRangeString(SiteRect& siteRect,
                                   const string& areaGroupName,
                                   const string& rangeString)
{
   bool gotOne(true);
   boost::smatch result;
   if (boost::regex_match(rangeString, result, areaRangePat)) {
      siteRect.areaGroupName = areaGroupName;
      siteRect.rangeString = rangeString;
      siteRect.siteType = result.str(1);
      transform(siteRect.siteType.begin(), siteRect.siteType.end(),
                siteRect.siteType.begin(), ::toupper);
      GetAreaRangeCorners(siteRect);
   } else if (boost::regex_match(rangeString, sitePat)) {
      string doubledSite(rangeString + ":" + rangeString);
      gotOne = ParseRangeString(siteRect,
                                areaGroupName,
                                doubledSite);
   } else {
      gotOne = false;
   }
   return gotOne;
}

// The following functions use a cache because the results don't
// change during a single run, computing the results requires a
// regular expression evaluation, and the functions may be called
// frequently. 

const boost::regex virtex4ArchPat("virtex4|qvirtex4|qrvirtex4|avirtex4");

// Note: This does not include the radiation hardened space-grade Virtex-5QV
// (device xqr5vfx130).
const boost::regex virtex5ArchPat("q?virtex5");

// Is the current device a member of the Virtex-4 family?
bool DeviceModel::IsVirtex4() const {
   static int cache(-1);
   if (cache == -1) {
      cache = boost::regex_match(GetArchitectureName(), virtex4ArchPat);
   }
   return cache;
}

// Is the current device a member of the Virtex-5 family?
bool DeviceModel::IsVirtex5() const {
   static int cache(-1);
   if (cache == -1) {
      cache = boost::regex_match(GetArchitectureName(), virtex5ArchPat);
   }
   return cache;
}

const boost::regex
spartan6ArchPat("spartan6|spartan6l|qspartan6l|qspartan6|aspartan6");

// Is the current device a member of the Spartan-6 family?
bool DeviceModel::IsSpartan6() const {
   static int cache(-1);
   if (cache == -1) {
      cache = boost::regex_match(GetArchitectureName(), spartan6ArchPat);
   }
   return cache;
}

const boost::regex sevenSeriesArchPat("[aq]?(?:virtex7|kintex7|artix7|zynq)l?");

// Is the current device a member of the 7 series, i.e. Artix-7,
// Kintex-7, Virtex-7, or Zynq-7?
bool DeviceModel::Is7Series() const {
   static int cache(-1);
   if (cache == -1) {
      cache = boost::regex_match(GetArchitectureName(), sevenSeriesArchPat);
   }
   return cache;
}

const boost::regex zynqArchPat("[aq]?(?:|zynq)l?");

bool DeviceModel::IsZynq() const {
   static int cache(-1);
   if (cache == -1) {
      cache = boost::regex_match(GetArchitectureName(), zynqArchPat);
   }
   return cache;
}
