// Copyright (C) 2006-2013 Xilinx, Inc.  All rights reserved
// $Id: PinGrid.c,v 1.29.206.4 2013/07/06 22:39:02 jcorbett Exp $
#include "PinGrid.h"

#include "Portability/export/Port_Type.h"
#include "Databases/Rf/export/Facade/Rf_Package.h"
#include "Databases/Rf/export/Facade/Rf_Cell.h"
#include "TUtilities/export/regex.hpp"
#include "Databases/Rf/export/Facade/Rf_Grid.h"
#include "GenericAlgorithms/Utilities/export/UtilCoord.h"

#include "GridMap.h"
#include "ArgParser.h"
#include "DeviceModel.h"

#include <sstream>
#include <iomanip>

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.29.206.4 $";

PinGrid::PinGrid(const DeviceModel* deviceModel)
   : bankFaultCount(0)
{
   Port::INT32 pkgCols(deviceModel->GetPackageColumnCount());
   Port::INT32 pkgRows(deviceModel->GetPackageRowCount());
   pkgPinGrid.reset(new GridMap<PinGroup>(pkgCols, pkgRows, pkgPins));

   Port::UINT32 dieCols(deviceModel->GetTileColumnCount());
   Port::UINT32 dieRows(deviceModel->GetTileRowCount());
   ioBufferGrid.reset(new GridMap<PinGroup>(dieCols, dieRows, ioBuffers));
}

PinGrid::~PinGrid() {
   for (std::vector<PinGroup*>::iterator iobIt(ioBuffers.begin());
        iobIt != ioBuffers.end();
        ++iobIt)
   {
      delete *iobIt;
   }
   for (std::vector<PinGroup*>::iterator ppIt(pkgPins.begin());
        ppIt != pkgPins.end();
        ++ppIt)
   {
      delete *ppIt;
   }
}

// R#check_package_pins, R#check_io_banks - Fill the
// package pin grid with data needed for checking package pin
// adjacency constraint violations, namely the pin name, the isolation
// group, the I/O bank, and, implicitly the grid coordinates.
void PinGrid::LabelPin(const Rf_PackagePin* pkgPin,
                       const string& netName,
                       const string& group)
{
   if (!pkgPin) {
      FatalInternal(39, "Invalid package pin for net " + netName
                    + " in group " + group + ".");
   }
   Rf_Cell ioBuffer(pkgPin->GetCell());
   if (!ioBuffer) {
      FatalInternal(43, "Unable to get I/O buffer from package pin on net "
                    + netName + ".");
   }
   string ioBufferName(ioBuffer.GetName());
   string pkgPinName(pkgPin->GetName());
   unsigned int px(pkgPin->GetColumn());
   unsigned int py(pkgPin->GetRow());
   PinGroup* p(new PinGroup(pkgPinName, netName, group, pkgPinFaults));

   // Note: GetTileColumn() and GetTileRow() return undefined results
   // for S6 pads, presumably because pads are at the site level, not
   // the tile level.
   Rf_BaseGrid* bg(ioBuffer.GetDevice()->GetBaseGrid());
   ASSERT(bg);
   Util_Point32 xy(bg->GetGridPoint(ioBuffer));
   ASSERT(0 <= xy.getx());
   ASSERT(0 <= xy.gety());
   unsigned int ix(static_cast<unsigned int>(xy.getx()));
   unsigned int iy(static_cast<unsigned int>(xy.gety()));
   Port::UINT32 bank(pkgPin->GetVccoBank());
   string bankStr(bank != UINT_MAX ? ToString(bank) : string("None"));

   PinGroup* d(new PinGroup(ioBufferName,
                            netName,
                            group,
                            ioBufferFaults,
                            bank));
   bool gotDuplicate(false);
   if (pkgPinGrid->At(px, py)) {
      Warn(108, "The package pin below was listed more than once.");
      gotDuplicate = true;
   }
   if (ioBufferGrid->At(ix, iy)) {
      Warn(109, "The I/O buffer below was listed more than once.");
      gotDuplicate = true;
   }
   if (ArgParser::verbose || gotDuplicate) {
      ostringstream out;
      out << setfill(' ')
          << "  "
          << setw(5) << pkgPinName
          << "("
          << setw(3) << px
          << ", "
          << setw(3) << py
          << ")"
          << setw(5) << bankStr
          << setw(16) << ioBufferName
          << " "
          << setw(16) << group
          << " "
          << netName;
      Report(out.str());
   }
   if (!gotDuplicate) {
      pkgPinGrid->MarkPoint(px, py, p);
      ioBufferGrid->MarkPoint(ix, iy, d);
   }
   ConsoleDot();
}

void PinGrid::CheckPinLocations() { // R#check_package_pins,
                                    // R#report_io_assignments
   pkgPinGrid->SweepAllAdjacent();
   ioBufferGrid->SweepVerticallyAdjacent();
}


// R#check_io_banks - IO buffers from different isolation groups
// cannot share a bank.
void PinGrid::CheckIoBanks() {
   // For each bank, collect one representative pin for each group that
   // occupies the bank. There should only be one per bank. If there
   // are several, then these will be reported later.
   // Note: Not all I/Os have an associated I/O bank. The constant
   // LNOINDEX (0xffffffff) is used to signify no I/O bank.
   for (std::vector<PinGroup*>::iterator iobIt(ioBuffers.begin());
        iobIt != ioBuffers.end();
        ++iobIt)
   {
      PinGroup* iob(*iobIt);
      if (iob &&
          iob->iobank != static_cast<int>(LNOINDEX) &&
          iob->group != "ignored" &&
          iob->group != "combined")
      {
         BankGroup::iterator bgIt(bankGroup.find(iob->iobank));
         if (bgIt == bankGroup.end()) {
            // Insert first representative.
            GroupPinMap gpm;
            gpm.insert(make_pair(iob->group, iob));
            bankGroup.insert(make_pair(iob->iobank, gpm));
         } else {
            GroupPinMap& gpm(bgIt->second);
            // Recall insert() returns a pair, the second is a Boolean that
            // indicates if something actually got inserted.
            bool didInsert(gpm.insert(make_pair(iob->group, iob)).second);
            if (didInsert) {
               ++bankFaultCount;
            }
         }
      }
   }
}

// R#report_errors, R#check_package_pins, R#check_io_banks,
// R#report_io_assignments
void PinGrid::ReportFaults() {

   // R#check_interconnect_routing, R#check_direct_routing,
   // R#check_logic_tiles
   Section("I/O Buffer Isolation Violations");

   if (ioBufferFaults.size()) {
      Report("Isolation Violation:122: No I/O buffers from distinct "
             "isolation\n"
             "groups shall be adjacent.");

      sort(ioBufferFaults.begin(), ioBufferFaults.end(),
           PinGroup::PinFaultLess());
      for (PinFaultIt pfIt(ioBufferFaults.begin());
           pfIt != ioBufferFaults.end();
           ++pfIt)
      {
         Report("\n  IOB " + pfIt->first->pinName
                + " on net " + pfIt->first->netName
                + " in group " + pfIt->first->group
                + " is adjacent\n"
                "  to " + pfIt->second->pinName
                + " on net " + pfIt->second->netName
                + " in group " + pfIt->second->group
                + ".");
      }
   } else {
      Report("  No I/O buffer isolation violations were found.");
   }

   // R#check_package_pins
   Section("Package Pin Isolation Violations");

   if (pkgPinFaults.size()) {
      Report("Isolation Violation:123: No pins on the package from distinct\n"
             "isolation groups shall be adjacent.");

      sort(pkgPinFaults.begin(), pkgPinFaults.end(),
           PinGroup::PinFaultLess());
      for (PinFaultIt pfIt(pkgPinFaults.begin());
           pfIt != pkgPinFaults.end();
           ++pfIt)
      {
         Report("\n  Pin " + pfIt->first->pinName
                + " on net " + pfIt->first->netName
                + " in group " + pfIt->first->group
                + " is adjacent\n"
                "  to pin " + pfIt->second->pinName
                + " on net " + pfIt->second->netName
                + " in group " + pfIt->second->group
                + ".");
      }
   } else {
      Report("  No package pin isolation violations were found.");
   }


   // R#check_io_banks
   Section("I/O Bank Isolation Violations");

   if (0 < bankFaultCount) {

      Report("Isolation Violation:124: No I/O buffers from distinct\n"
             "isolation groups shall share an I/O bank.");

      for (BankGroupIt bgIt(bankGroup.begin());
           bgIt != bankGroup.end();
           ++bgIt)
      {
         GroupPinMap& gpm(bgIt->second);
         if (1 < gpm.size()) {  // R#check_io_banks
            // loop over gpm contents
            Report("\n  I/O Bank " + ToString(bgIt->first)
                   + " contains pins from " + ToString(gpm.size())
                   + " groups.");
            for (GroupPinMapIt gpmIt(gpm.begin());
                 gpmIt != gpm.end();
                 ++gpmIt)
            {
               Report("    Group: " + gpmIt->first
                      + ", e.g. " + gpmIt->second->pinName);
            }
         }
      }
   } else {
      Report("  No I/O bank violations were found.");
   }

   Summary("\n  I/O Isolation\n");

   if (!pkgPins.empty()) {
      Summary("    I/O Buffer Isolation Violations: "
              + ToString(ioBufferFaults.size())
              + "\n    Package Pin Isolation Violations: "
              + ToString(pkgPinFaults.size())
              + "\n    Bank Isolation Violations: "
              + ToString(bankFaultCount));
   } else {
      Summary("    I/O analysis was not requested.");
   }
}

int PinGrid::GetStatus() {
  return (ioBufferFaults.size() + pkgPinFaults.size() + bankFaultCount) != 0;
}

int PinGrid::GetErrorCount() {
  return ioBufferFaults.size() + pkgPinFaults.size() + bankFaultCount;
}
