// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: Tile.c,v 1.19.206.17 2013/07/06 22:34:58 jcorbett Exp $
#include "Tile.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.19.206.17 $";

// R#check_irs_separation, R#draw_fence
void Tile::SetCell(Rf_Cell cell) {
   this->cell = cell;
}

// R#check_irs_separation, R#draw_fence
Rf_Cell Tile::GetCell() {
   return cell;
}

// Clock buffers are different than other logic sites. They connect
// directly to interconnects across wide areas. They do not normally
// transmit critical program information. If they are treated like
// other logic sites, spurious isolation violation are reported.
//
// This pattern is used to ignore clock buffers and related resources.
const boost::regex
clockPat("/?(?:BUFGCTRL|BUFGMUX|BUFH|BUFIO|"
         "BUFMR|BUFPLL|BUFR|DCI|IDELAYCTRL).*");

// Add the name of a logic component to this tile.
void Tile::AddCompName(const string compName) {
   compSet.insert(compName);
}

// R#check_irs_separation, R#draw_fence
void Tile::AddLogicSite(Rf_Cell logicSite) {
   ASSERT(logicSite);
   string compName(logicSite.GetName());
   if (!boost::regex_match(compName, clockPat)) {
      logicSiteSet.insert(logicSite);
   }
}

// R#check_irs_separation, R#draw_fence
Rf_Cell Tile::GetLogicSite() {
   ASSERT(logicSiteSet.size() == 1);
   return *logicSiteSet.begin();
}

// Does this tile contain a site of the indicated type?
bool Tile::HasSiteType(const string& siteType) {
   bool result(false);
   for (DeviceModel::CellSetIt lsIt(logicSiteSet.begin());
        lsIt != logicSiteSet.end();
        ++lsIt)
   {
      Rf_Cell site(*lsIt);
      string typeStr(site.GetCellDef().GetName());
      if (typeStr == siteType) {
         result = true;
         break;
      }
   }
   return result;
}

// R#check_irs_separation, R#draw_fence
void Tile::SetGroupSet(Data::GroupSet* groupSet) {
   this->groupSet = groupSet;
}

// R#check_irs_separation, R#draw_fence
Data::GroupSet* Tile::GetGroupSet() const {
   return groupSet;
}

// To identify gaps in the fence composed one or more invalid fence
// tiles*, isolated neighbors are listed within the tile. The set of
// isolation groups of the neighbors is maintained as new tainting
// neighbors are added.
//
// * - Isolation violations can also happen as a result of directly
// adjacent tiles from distinct isolation groups or individual tiles
// containing logic/routing from multiple isolation groups.
Data::GroupSet* Tile::GetTaintGroupSet() const {
   return taintGroupSet;
}

// R#check_irs_separation, R#draw_fence
bool Tile::Equal(const Tile& tile) const {
   return groupSet == tile.GetGroupSet();
}

// R#draw_fence
void Tile::SetHasBusMacro() {
   hasBusMacro = true;
}

// R#draw_fence
bool Tile::HasBusMacro() const {
   return hasBusMacro;
}

// R#draw_fence
void Tile::SetHasGlobalNet() {
   hasGlobalNet = true;
}

// R#draw_fence
bool Tile::HasGlobalNet() const {
   return hasGlobalNet;
}

// R#draw_fence
void Tile::SetHasIsolatedNet() {
   hasIsolatedNet = true;
}

// R#draw_fence
bool Tile::HasIsolatedNet() const {
   return hasIsolatedNet;
}

// R#check_irs_separation, R#draw_fence
void Tile::SetHasBadNeighbor() {
   hasBadNeighbor = true;
}

// R#check_irs_separation, R#draw_fence
bool Tile::HasBadNeighbor() const {
   return hasBadNeighbor;
}

// R#check_irs_separation, R#draw_fence - Is there a comp in this
// tile?
bool Tile::HasProgramming() const {
   return !compSet.empty();
}

// R#check_irs_separation, R#draw_fence
// Record terminals (aka programmable interconnect points or PIPs)
// indexed by net.
void Tile::AddTerminal(XCad_CadNetwork* net, Rf_Wire wire) {
   irsPips[net].insert(wire);
}

// R#check_irs_separation, R#draw_fence
void Tile::ClearTerminals() {
   irsPips.clear();
}

// R#check_irs_separation, R#draw_fence
bool Tile::HasTerminal() const {
   return !irsPips.empty();
}

// The name of this tile is the name of its cell.
const string Tile::GetName() const {
   ASSERT(cell);
   return string(cell.GetName());
}

const string site("[_A-Z0-9]+");
const string num("(\\d+)");       // a non-negative integer
const boost::regex cellXYPat(site + "_X" + num + "Y" + num,
                             boost::regex::normal);

// Derive a string from the tile name of the form
// "x, y" to reduce label overlaps.
const string Tile::GetLabel() const {
   ASSERT(cell);
   boost::smatch result;
   const string name(GetName());
   bool gotMatch(boost::regex_match(name, result, cellXYPat));
   
   // Set xyLabel to "x, y" if the pattern matches; fall back
   // to the cell name otherwise.
   const string
      xyLabel(gotMatch
              ? result.str(1) + ", " + result.str(2)
              : name);
   return xyLabel;
}

// Add the description of a net that enters this tile.
void Tile::AddNetDescription(const string description) {
   netDescriptions.insert(description);
}

// Record the net names for the purpose of counting the nets used by a
// collection of tiles.
//
// This routine stores the names rather than the pointers because some
// user nets are represented internally with multiple XCad_CadNetwork
// instances. E.g. nets associated with IBUF and OBUF sites.
void Tile::AddNet(XCad_CadNetwork* net) {
   string netName(net->GetUserHierName());
   netSet.insert(netName);
}

// Add the nets from this tile to the given set.
void Tile::CollectNets(Classifier::NameSet& netSet) {
   netSet.insert(this->netSet.begin(), this->netSet.end());
}

// R#check_irs_separation, R#draw_fence, R#check_interconnect_routing -
// There are two ways for a tile to have bad content:
//    It contains nets from multiple isolation groups; or
//    It is needed as a fence tile, but is invalid because
//        it contains programming or bad terminals.
//
// Note: It is only valid to call this method after
// TileChecker::SweepAllAdjacent() has finished.
bool Tile::HasBadContent() const {
   return groupSet && 1 < groupSet->size();
}

// A isolated tile has an isolation violation if any of
// the following:
//    1. It has a neighbor from another isolation group, or
//    2. It has content from another isolation group.
bool Tile::HasViolation() const {
   return HasBadNeighbor() || HasBadContent();
}

// R#draw_fence, R#check_irs_separation, R#check_interconnect_routing -
// For a tile to be used as part of the fence, all the following must
// be true:
//    It must not contain any isolated nets.
//    It must not contain any logic (neither isolated nor global).
//    It must not use any PIPs associated with non-clock routing.
bool Tile::IsValidFenceTile() const {
   return !(HasIsolatedNet() ||
            HasProgramming() ||
            HasTerminal() ||
            HasGroups() ||
            HasTaints());
}

// R#draw_fence, R#check_irs_separation, R#check_interconnect_routing -
// This tile has at least one group.
// 
// Note: It is valid to call this after TileChecker::MapNetsToTiles()
// or TileChecker::MapSitesToTiles.
bool Tile::HasGroups() const {
   return groupSet && !groupSet->empty();
}

// Add a neighboring supersite to the set of supersites adjacent to
// this tile. Of all the supersites, only the ones considered big dead
// shorts are added. This set is used to propagate isolation group
// information across those supersites.
void Tile::AddNeighbor(Tile* adjoined) {
   neighborSet.insert(adjoined);
}

TileSet& Tile::GetNeighbors() {
   return neighborSet;
}

// Count the number of distinct groups represented by the isolated
// direct neighbors of this tile.
size_t Tile::CountNeighborGroups() {
   Data::GroupSet* gs(0);
   for (TileSetIt tsIt(neighborSet.begin());
        tsIt != neighborSet.end();
        ++tsIt)
   {
      Tile* neighbor(*tsIt);
      if (neighbor->HasGroups()) {
         gs = groupSetFamily.FindOrMakeUnion(gs, neighbor->GetGroupSet());
      }
   }
   size_t result(0);
   if (gs) {
      result = gs->size();
   }
   return result;
}

// Are there isolated neighbors or indirect neighbors that might pass
// influence through this tile?
bool Tile::HasTaints() const {
   return !taintSet.empty();
}

// The set of neighbors and indirect neighbors (isolated tiles or other
// tainted tiles) that cannot be ruled out as influences potentially
// violating isolation.
const TileSet& Tile::GetTaints() const {
   return taintSet;
}

// Return the number of isolation groups from which this tile is not
// isolated. For correctly isolated designs, this should never exceed
// one.
size_t Tile::CountTaints() {
   return taintGroupSet ? taintGroupSet->size() : 0;
}

// Is this tile not isolated from multiple isolation groups?
bool Tile::HasMultipleTaints() {
   return 1 < CountTaints();
}

// R#report_interregion_signals - Generate a textual
// representation of a set. If the separator is not
// specified, a comma-separated list is produced.
const string SetToString(const Data::GroupSet* gs,
                         const string separator)
{
   string result;
   string sep;
   if (gs) {
      for (Data::GroupSet::const_iterator i(gs->begin());
           i != gs->end();
           ++i)
      {
         if (*i == "ignored") {
            result += sep + 'i';
         } else {
            result += sep + *i;
         }
         sep = separator;
      }
   }
   if (result.empty()) {
      result = "(empty)";
   }
   return result;
}

// Generate a string describing this tile. Used to aid debugging.
// This is used during SVG diagram generation to add comments with
// detail about the disposition of a tile that is unusual in some way,
// i.e. it contains an isolation violation.
const string Tile::Describe() const {
   string code("Tile: " + GetName() + " Properties: ");
   if (HasBadNeighbor()) code += " bad_neighbor";
   if (HasTerminal())    code += " irs_with_pips";
   if (HasProgramming()) code += " has_programming ";
   if (HasBadContent())  code += " irs_not_isolated";

   if (HasGlobalNet())   code += " has_global";
   if (HasIsolatedNet()) code += " has_isolated_net";
   if (!logicSiteSet.empty()) {
      code += " Logic: {";
      string sep;
      for (DeviceModel::CellSet::const_iterator lsIt(logicSiteSet.begin());
           lsIt != logicSiteSet.end();
           ++lsIt)
      {
         Rf_Cell site(*lsIt);
         string siteName(site.GetName());
         code += sep + siteName;
         sep = ",";
      }
      code += "}";
   }
   if (HasGroups())      code += " Groups: {" + SetToString(groupSet, ",") + "}";
   if (HasTaints()) {
      string taintDescription;
      TileSet taintSet(GetTaints());
      string sep;
      for (TileSetIt tsIt(taintSet.begin());
           tsIt != taintSet.end();
           ++tsIt)
      {
         Tile* tile(*tsIt);
         taintDescription += sep + tile->GetName();
         if (tile->HasGroups()) {
            taintDescription += "[" + SetToString(tile->GetGroupSet(), ",") + "]";
         }
         sep = ",";
      }
      string taintGroups(SetToString(GetTaintGroupSet(), ","));
      code += " Taints: {" + taintDescription + "}";
      code += " TaintGroups: {" + taintGroups + "}";
   }
   return code;
}

// Return a string with first line only of each string in
// netDescriptions indented four spaces. Suppresses duplicates and
// sorts. This was an after thought when we realized that it is not
// always helpful to print all the detail of the nets in a tile.
const string Tile::GetNetsBrief() const {
   Classifier::NameSet result;
   for (Classifier::NameSetIt nsIt(netDescriptions.begin());
        nsIt != netDescriptions.end();
        ++nsIt)
   {
      string full(*nsIt);
      size_t eol(full.find('\n'));
      ASSERT(eol != string::npos);
      ASSERT(eol);
      result.insert(full.substr(0, eol));
   }
   return "    " + ToString(result, "\n    ");
}

// Mark this as tainted by tile and update the group set of all the
// tiles that potentially influence this one. A true return value
// indicates that the set of tainting groups changed.
//
// Note: taints only make sense for NCD mode where there is actual
// data about routing and logic between isolated regions.
bool Tile::AddTaint(Tile* tile) {
   bool result(false);
   if (!isUcfMode) {
      bool didInsert(taintSet.insert(tile).second);
      size_t before(0);
      if (taintGroupSet) {
         before = taintGroupSet->size();
      }
      if (didInsert) {
         taintGroupSet =
            groupSetFamily.FindOrMakeUnion(taintGroupSet,
                                           tile->GetTaintGroupSet());
         taintGroupSet =
            groupSetFamily.FindOrMakeUnion(taintGroupSet,
                                           tile->GetGroupSet());

      }
      size_t after(0);
      if (taintGroupSet) {
         after = taintGroupSet->size();
      }
      result = before != after;
   }
   return result;
}

// Propagation is used to handle chains of invalid fence tiles
// connecting isolated regions.
//
// Should isolation groups propagate from this to adjoined?
// The source has something to propagate.
// The destination isn't a valid fence tile.
bool Tile::ShouldPropagateTo(Tile* adjoined) const {
   bool result(!adjoined->HasGroups() &&
               (HasGroups() || HasTaints()) &&
               !adjoined->IsValidFenceTile());
   return result;
}

// Would it violate isolation if this tile influenced the given tile?
// Need to check all combinations of direct isolation groups and
// other influences (taints).
bool Tile::IsViolation(Tile* a) const {
   return
      (HasGroups() && a->HasGroups() && !Equal(*a)) ||
      (HasTaints() && a->HasTaints() && taintGroupSet != a->taintGroupSet) ||
      (HasTaints() && a->HasGroups() && taintGroupSet != a->GetGroupSet()) ||
      (HasGroups() && a->HasTaints() && GetGroupSet() != a->taintGroupSet);
}

// R#check_irs_separation, R#draw_fence, R#check_interconnect_routing -
// Isolated routing of one group must not be in an adjacent tile to
// isolated routing of another group. A tile is needed as a fence tile
// if it has at least two neighbors that need a fence between
// them. E.g. one neighbor is red or not isolated from red and another
// neighbor is black or not isolated from black.
//
// If a tile is needed as a fence tile, but contains something that
// renders it invalid as a fence tile, adjacency information is
// propagated through the tile via a tainting mechanism.
//
// A tile is isolated if all of its neighbors are either in the same
// isolation group or valid fence tiles.
void Tile::Visit(Tile* adjoined) {
   if (sweepState != 2) {

      // Spread the isolation groups from isolated tiles or invalid fence
      // tiles to other invalid fence tile. This process is repeated in
      // passes over the array of interconnect tiles until no further
      // changes are made.
      //
      // Note: Visit is only invoked for pairs in which this precedes
      // adjoined in the tile array. Therefore, to spread all directions,
      // we must spread from this to adjoined and from adjoined to this.
      if (ShouldPropagateTo(adjoined)) {
         bool gainedGroups(adjoined->AddTaint(this));
         if (gainedGroups) sweepState = 1;
      } else if (adjoined->ShouldPropagateTo(this)) {
         bool gainedGroups(this->AddTaint(adjoined));
         if (gainedGroups) sweepState = 1;
      }
   } else {
      // See if this and adjoined create an isolation violation.
      if (IsViolation(adjoined)) {
         // Record a fault.
         if (HasIsolatedNet() && adjoined->HasIsolatedNet()) {
            netAdjacencyFaults.push_back(make_pair(this, adjoined));
         }
         // The bad neighbor bit will generate fence PIPs violations.
         SetHasBadNeighbor();
         adjoined->SetHasBadNeighbor();
      
         // If there's no fence, is the grass greener on the other side of it?
      }
   }
}
